import argparse
from typing import Any

from ..serializers.jsonserializer import JSONSerializer
from ..serializers.xmlserializer import XMLSerializer


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('file_from')
    parser.add_argument('file_to')
    parser.add_argument('format_from')
    parser.add_argument('format_to')

    args = parser.parse_args()
    file_from, file_to, format_from, format_to = (
        args.file_from,
        args.file_to,
        args.format_from,
        args.format_to
    )

    format_mapping: dict[str, Any] = {
        'json': JSONSerializer(),
        'xml': XMLSerializer()
    }

    with open(file_from, 'r') as ff, open(file_to, 'w+') as ft:
        format_from = format_mapping[format_from]
        format_to = format_mapping[format_to]

        format_to.dump(format_from.load(ff), ft)


if __name__ == "__main__":
    main()
    


from __future__ import annotations

import re

from types import FunctionType, MethodType, CodeType, ModuleType, \
    BuiltinMethodType, BuiltinFunctionType, CellType, GeneratorType
from typing import Any, Collection, Iterable

from ..helpers.constants import IGNORED_FIELDS, IGNORED_FIELD_TYPES, TYPE_MAPPING


def get_items(obj) -> dict[str, Any]:
    if isinstance(obj, (BuiltinFunctionType, BuiltinMethodType)):
        return {}

    if isinstance(obj, dict):
        return obj

    elif isinstance(obj, Collection):
        return dict(enumerate(obj))

    elif isinstance(obj, CodeType):
        return {
            "argcount": obj.co_argcount,
            "posonlyargcount": obj.co_posonlyargcount,
            "kwonlyargcount": obj.co_kwonlyargcount,
            "nlocals": obj.co_nlocals,
            "stacksize": obj.co_stacksize,
            "flags": obj.co_flags,
            "code": obj.co_code,
            "consts": obj.co_consts,
            "names": obj.co_names,
            "varnames": obj.co_varnames,
            "filename": obj.co_filename,
            "name": obj.co_name,
            "firstlineno": obj.co_firstlineno,
            "lnotab": obj.co_lnotab,
            "freevars": obj.co_freevars,
            "cellvars": obj.co_cellvars,
        }
    elif isinstance(obj, property):
        return {
            "getter": obj.fget,
            "setter": obj.fset,
            "deleter": obj.fdel,
        }

    elif isinstance(obj, GeneratorType):
        a = []
        for i in obj:
            a.append(i)
        return{
            "values": a
        }

    elif isinstance(obj, FunctionType):
        if obj.__closure__ and "__class__" in obj.__code__.co_freevars:
            closure = ([... for _ in obj.__closure__])
        elif obj.__closure__:
            closure = ([cell.cell_contents for cell in obj.__closure__])
        else:
            closure = None

        return {
            "argcount": obj.__code__.co_argcount,
            "posonlyargcount": obj.__code__.co_posonlyargcount,
            "kwonlyargcount": obj.__code__.co_kwonlyargcount,
            "nlocals": obj.__code__.co_nlocals,
            "stacksize": obj.__code__.co_stacksize,
            "flags": obj.__code__.co_flags,
            "code": obj.__code__.co_code,
            "consts": obj.__code__.co_consts,
            "names": obj.__code__.co_names,
            "varnames": obj.__code__.co_varnames,
            "filename": obj.__code__.co_filename,
            "name": obj.__code__.co_name,
            "firstlineno": obj.__code__.co_firstlineno,
            "lnotab": obj.__code__.co_lnotab,
            "freevars": obj.__code__.co_freevars,
            "cellvars": obj.__code__.co_cellvars,
            "globals": {
                k: obj.__globals__[k]
                for k in (
                        set(
                            k for k, v in obj.__globals__.items()
                            if isinstance(v, ModuleType)
                        ) |
                        set(obj.__globals__) &
                        set(obj.__code__.co_names) -
                        {obj.__name__}
                )
            },
            "closure": closure,
            "qualname": obj.__qualname__
        }

    elif isinstance(obj, MethodType):
        return {
            "__func__": obj.__func__,
            "__self__": obj.__self__
        }

    elif issubclass(type(obj), type):
        return {
            'name': obj.__name__,
            'mro': tuple(obj.mro()[1:-1]),
            'attrs': {
                k: v for k, v in obj.__dict__.items()
                if (
                        k not in IGNORED_FIELDS and
                        type(v) not in IGNORED_FIELD_TYPES
                )
            }
        }

    elif issubclass(type(obj), ModuleType):
        return {'name': obj.__name__}

    elif isinstance(obj, staticmethod):
        return get_items(obj.__func__)

    elif isinstance(obj, classmethod):
        return get_items(obj.__func__)

    else:
        return {
            'class': obj.__class__,
            'attrs': {
                k: v for k, v in obj.__dict__.items()
                if (
                        k not in IGNORED_FIELDS and
                        type(k) not in IGNORED_FIELD_TYPES
                )
            }
        }


def get_key(value, obj: dict) -> int:
    return [key for key in obj if obj[key] == value][0]


def to_number(s: str) -> int | float | complex | None:
    for num_type in (int, float, complex):
        try:
            return num_type(s)
        except (ValueError, TypeError):
            pass


def create_object(obj_type: type, obj_data):
    if issubclass(obj_type, dict):
        return obj_data

    elif issubclass(obj_type, GeneratorType):
        it = iter(obj_data.get("values"))

        def generator_func():
            yield from it

        return generator_func()

    elif issubclass(obj_type, Iterable):
        return obj_type(obj_data.values())

    elif issubclass(obj_type, CodeType):
        return CodeType(*list(obj_data.values()))

    elif issubclass(obj_type, FunctionType):
        if obj_data.get('closure'):
            closure = tuple([CellType(x) for x in obj_data.get('closure')])
        elif obj_data.get('closure') and '__class__' in obj_data.get('freevars'):
            closure = tuple([CellType(...) for _ in obj_data.get('closure')])
        else:
            closure = tuple()

        obj = FunctionType(
            code=CodeType(*list(obj_data.values())[:16]),
            globals=obj_data.get('globals'),
            name=obj_data['name'],
            closure=closure
        )
        obj.__qualname__ = obj_data.get('qualname')
        obj.__globals__[obj.__name__] = obj

        return obj

    elif issubclass(obj_type, property):
        return property(fget=obj_data.get("getter"),
                        fdel=obj_data.get("deleter"),
                        fset=obj_data.get("setter"))



    elif issubclass(obj_type, MethodType):
        return MethodType(
            obj_data.get('__func__'),
            obj_data.get('__self__'),
        )

    elif issubclass(obj_type, (staticmethod, classmethod)):
        return create_object(FunctionType, obj_data)

    elif issubclass(obj_type, type):
        obj = type(obj_data.get('name'), obj_data.get('mro'), obj_data.get('attrs'))

        try:
            obj.__init__.__closure__[0].cell_contents = obj
        except (AttributeError, IndexError):
            ...

        return obj

    elif issubclass(obj_type, ModuleType):
        return __import__(obj_data.get('name'))

    # if obj_data[TYPE] == ITERATOR_TYPE:
    #     return iter(create_object(obj_type, item) for item in obj_data[VALUE])

    # elif issubclass(obj_type, type(iter)):
    #     return iter(create_object(obj_type, item) for item in obj_data[VALUE])

    else:
        obj = object.__new__(obj_data.get('class'))
        obj.__dict__ = obj_data.get('attrs')
        return obj


def type_from_str(s: str, pattern: str) -> type:
    if not re.search(pattern, s):
        return type(None)

    return TYPE_MAPPING[re.search(pattern, s).group(1)]


def is_iterable(obj):
    return hasattr(obj, '__iter__') and hasattr(obj, '__next__') and callable(obj.__iter__)



import re
from typing import Iterator

from ..helpers.constants import JSON, BOOL_TYPE, TYPE_MAPPING, JSON_TYPE, PRIMITIVE_TYPES
from ..helpers.functions import get_items, get_key, to_number, create_object, type_from_str


class JSONSerializer:
    def dumps(self, obj) -> str:
        if type(obj) == str:
            return f'"{obj}"'
        if type(obj) in (int, float, complex):
            return str(obj)
        if type(obj) in [bool, type(None)]:
            return BOOL_TYPE[obj]

        return JSON.format(
            type=type(obj) if type(obj) in TYPE_MAPPING.values() else object,
            id=id(obj),
            items=self.__load_to_json(get_items(obj))
        )

    def loads(self, json: str):
        if not len(json):
            return

        if json == ' ':
            return ...
        if json.startswith('"'):
            return json.strip('"')
        if json in BOOL_TYPE.values():
            return get_key(json, BOOL_TYPE)
        if to_number(json) is not None:
            return to_number(json)

        return create_object(
            type_from_str(json, JSON_TYPE),
            self.__load_from_json(json)
        )

    def __load_to_json(self, obj: dict) -> str:
        json_format = ""

        for k, v in obj.items():
            if type(v) in PRIMITIVE_TYPES:
                json_format += f"\t{self.dumps(k)}: {self.dumps(v)},\n"
                continue

            json_format += f"\t{self.dumps(k)}: {{\n"

            for line in self.dumps(v).split("\n")[1:]:
                json_format += f"\t{line}\n"

        return json_format

    def __load_from_json(self, template: str) -> dict:
        obj: dict = {}
        lines: list[str] = template.split("\n")
        it: Iterator[str] = enumerate(lines)

        for i, line in it:
            if not re.search(r'\s*(.+):\s*([^,]*)', line):
                continue

            key, value = re.search(r'\s*(.+):\s*([^,]*)', line).groups()

            if value != "{":
                obj[self.loads(key)] = self.loads(value)

            elif value == "{" and "<class" not in key:
                brackets = 1
                start = i + 1

                while brackets and i < len(lines) - 1:
                    i, line = next(it, None)
                    brackets += ("{" in lines[i]) - ("}" in lines[i])

                obj[self.loads(key)] = self.loads('\n'.join(lines[start:i]))

        return obj

    def dump(self, obj, file_to) -> None:
        file_to.write(self.dumps(obj))

    def load(self, file_from):
        return self.loads(file_from.read())
        # return )



import re
from typing import Iterator

from ..helpers.constants import PRIMITIVE_TYPES, TYPE_MAPPING, XML, XML_ITEM, XML_PRIMITIVE, XML_TYPE
from ..helpers.functions import get_items, get_key, type_from_str, create_object


class XMLSerializer:
    def dumps(self, obj) -> str:
        if type(obj) in PRIMITIVE_TYPES:
            obj_type = get_key(type(obj), TYPE_MAPPING)
            return f'<primitive type="{obj_type}">{obj}</primitive>'

        return XML.format(
            type=get_key(type(obj), TYPE_MAPPING) if type(obj) in TYPE_MAPPING.values() else "object",
            id=id(obj),
            items=self.__load_to_xml(get_items(obj))
        )

    def loads(self, xml):
        if not len(xml):
            return

        if "primitive" in xml.split("\n")[0]:
            obj_data = re.search(
                XML_PRIMITIVE.format(
                    type="\w+",
                    obj="(.*)"
                ), xml).group(1)
            obj_type = type_from_str(
                s=xml.split("\n")[0],
                pattern=XML_TYPE
            )

            if obj_type == type(None):
                return None

            if obj_type == bool:
                return obj_data == "True"

            if obj_type == type(Ellipsis):
                return ...

            return obj_type(obj_data)

        return create_object(
            type_from_str(xml, XML_TYPE),
            self.__load_from_xml(xml)
        )

    def __load_to_xml(self, obj: dict) -> str:
        xml_format = ""

        for k, v in obj.items():
            xml_format += f"\t{XML_ITEM.format(key=self.dumps(k), value=self.dumps(v))}"
        return xml_format

    def __load_from_xml(self, template: str) -> dict:
        obj: dict = {}
        lines: list[str] = template.split("\n")
        it: Iterator[str] = enumerate(lines)

        for i, line in it:
            if "<item>" == line.strip("\t\n "):
                item = self.__get_tag("item", lines[i + 1:])
                key = self.__get_tag("key", item[1:])
                value = self.__get_tag("value", item[len(key) + 2:])

                obj[self.loads("\n".join(key[:-1]))] = self.loads("\n".join(value[:-1]))

                [next(it, None) for _ in range(len(item))]

        return obj

    def __get_tag(self, tagname: str, lines) -> str:
        counter = 1
        it = enumerate(lines)

        for i, line in it:
            if not counter:
                return lines[:i]

            counter += bool(re.search(rf"<{tagname}.*>", line.strip("\t\n ")))
            counter -= bool(re.search(rf"</{tagname}>", line.strip("\t\n ")))

    def dump(self, obj, file_to) -> None:
        file_to.write(self.dumps(obj))

    def load(self, file_from):
        return self.loads(file_from.read())



from setuptools import setup

setup(name='serjx',
      version='1.0',
      description='Serializer',
      packages=['Lab_3', 'Lab_3/helpers', 'Lab_3/serializers'],
      author_email='tatarinov.slava222@gmail.com',
      zip_safe=False)



from types import LambdaType, FunctionType, MethodType, CodeType,\
    ModuleType, CellType, GeneratorType
from typing import Final

JSON_TYPE: str = r"<class '(\w\S+)'>_"
JSON = ('''{{
    "{type}_{id:x}": {{
    {items}
    }}
}}''')


XML_TYPE: str = r'type="(\w+)"'
XML = ('''
<object type="{type}" id="{id:x}">
{items}
</object>
''')
XML_ITEM = ('''
<item>
    <key>
        {key}
    </key>
    <value>
        {value}
    </value>
</item>
''')
XML_PRIMITIVE = '<primitive type="{type}">{obj}</primitive>'

PRIMITIVE_TYPES: tuple = (int, float, complex, str, bool, type(None))

BOOL_TYPE: dict[bool, str] = {
        None: 'none',
        True: 'true',
        False: 'false'
    }

TYPE_MAPPING = {
    'int': int,
    'float': float,
    'complex': complex,
    'str': str,
    'bool': bool,
    'NoneType': type(None),
    'bytes': bytes,
    'list': list,
    'tuple': tuple,
    'set': set,
    'dict': dict,
    'code': CodeType,
    'cell': CellType,
    'function': FunctionType,
    'lambda': LambdaType,
    'method': MethodType,
    'staticmethod': staticmethod,
    'classmethod': classmethod,
    'type': type,
    'module': ModuleType,
    'object': object,
    'property': property,
    'generator': GeneratorType
}

IGNORED_FIELDS: set[str] = {
        '__weakref__',
        '__subclasshook__',
        '__dict__',
        '__doc__'
}
IGNORED_FIELD_TYPES: set[str] = {
    'BuiltinFunctionType', 'BuiltinMethodType',
    'WrapperDescriptorType', 'MethodDescriptorType',
    'MappingProxyType', 'GetSetDescriptorType',
    'MemberDescriptorType'
}

TYPE: Final[str] = "type"
ITERATOR_TYPE: Final[str] = "iterator"
VALUE: Final[str] = "value"

